@@ -343,7 +343,7 @@ SDL_Surface *IMG_LoadJPG_RW(SDL_RWops *src)
     Sint64 start;
     struct jpeg_decompress_struct cinfo;
     JSAMPROW rowptr[1];
-    SDL_Surface *volatile surface = NULL;
+    SDL_Surface *surface = NULL;
     struct my_error_mgr jerr;
 
     if ( !src ) {
@@ -477,71 +477,81 @@ static void jpeg_SDL_RW_dest(j_compress_ptr cinfo, SDL_RWops *ctx)
     dest->pub.free_in_buffer = OUTPUT_BUFFER_SIZE;
 }
 
-static int IMG_SaveJPG_RW_jpeglib(SDL_Surface *surface, SDL_RWops *dst, int freedst, int quality)
+struct savejpeg_vars
+{
+    struct jpeg_compress_struct cinfo;
+    struct my_error_mgr jerr;
+    Sint64 original_offset;
+};
+
+static int JPEG_SaveJPEG_RW(struct savejpeg_vars *vars, SDL_Surface *jpeg_surface, SDL_RWops *dst, int quality)
+{
+    /* Create a compression structure and load the JPEG header */
+    vars->cinfo.err = lib.jpeg_std_error(&vars->jerr.errmgr);
+    vars->jerr.errmgr.error_exit = my_error_exit;
+    vars->jerr.errmgr.output_message = output_no_message;
+    vars->original_offset = SDL_RWtell(dst);
+
+    if(setjmp(vars->jerr.escape)) {
+        /* If we get here, libjpeg found an error */
+        lib.jpeg_destroy_compress(&vars->cinfo);
+        SDL_RWseek(dst, vars->original_offset, RW_SEEK_SET);
+        return IMG_SetError("Error saving JPEG with libjpeg");
+    }
+
+    lib.jpeg_create_compress(&vars->cinfo);
+    jpeg_SDL_RW_dest(&vars->cinfo, dst);
+
+    vars->cinfo.image_width = jpeg_surface->w;
+    vars->cinfo.image_height = jpeg_surface->h;
+    vars->cinfo.in_color_space = JCS_RGB;
+    vars->cinfo.input_components = 3;
+
+    lib.jpeg_set_defaults(&vars->cinfo);
+    lib.jpeg_set_quality(&vars->cinfo, quality, TRUE);
+    lib.jpeg_start_compress(&vars->cinfo, TRUE);
+
+    while (vars->cinfo.next_scanline < vars->cinfo.image_height) {
+        JSAMPROW row_pointer[1];
+        int offset = vars->cinfo.next_scanline * jpeg_surface->pitch;
+
+        row_pointer[0] = ((Uint8*)jpeg_surface->pixels) + offset;
+        lib.jpeg_write_scanlines(&vars->cinfo, row_pointer, 1);
+    }
+
+    lib.jpeg_finish_compress(&vars->cinfo);
+    lib.jpeg_destroy_compress(&vars->cinfo);
+    return 0;
+}
+
+static int IMG_SaveJPG_RW_jpeglib(SDL_Surface *surface, SDL_RWops *dst, int quality)
 {
     /* The JPEG library reads bytes in R,G,B order, so this is the right
      * encoding for either endianness */
+    struct savejpeg_vars vars;
     static const Uint32 jpg_format = SDL_PIXELFORMAT_RGB24;
-    struct jpeg_compress_struct cinfo;
-    struct my_error_mgr jerr;
-    JSAMPROW row_pointer[1];
     SDL_Surface* jpeg_surface = surface;
-    int result = -1;
-
-    if (!dst) {
-        IMG_SetError("Passed NULL dst");
-        goto done;
-    }
+    int ret;
 
     if (!IMG_Init(IMG_INIT_JPG)) {
-        goto done;
+        return -1;
     }
 
     /* Convert surface to format we can save */
     if (surface->format->format != jpg_format) {
         jpeg_surface = SDL_ConvertSurfaceFormat(surface, jpg_format, 0);
         if (!jpeg_surface) {
-            goto done;
+            return -1;
         }
     }
 
-    /* Create a decompression structure and load the JPEG header */
-    cinfo.err = lib.jpeg_std_error(&jerr.errmgr);
-    jerr.errmgr.error_exit = my_error_exit;
-    jerr.errmgr.output_message = output_no_message;
-
-    lib.jpeg_create_compress(&cinfo);
-    jpeg_SDL_RW_dest(&cinfo, dst);
-
-    cinfo.image_width = jpeg_surface->w;
-    cinfo.image_height = jpeg_surface->h;
-    cinfo.in_color_space = JCS_RGB;
-    cinfo.input_components = 3;
-
-    lib.jpeg_set_defaults(&cinfo);
-    lib.jpeg_set_quality(&cinfo, quality, TRUE);
-    lib.jpeg_start_compress(&cinfo, TRUE);
-
-    while (cinfo.next_scanline < cinfo.image_height) {
-        int offset = cinfo.next_scanline * jpeg_surface->pitch;
-        row_pointer[0] = ((Uint8*)jpeg_surface->pixels) + offset;
-        lib.jpeg_write_scanlines(&cinfo, row_pointer, 1);
-    }
-
-    lib.jpeg_finish_compress(&cinfo);
-    lib.jpeg_destroy_compress(&cinfo);
+    SDL_zero(vars);
+    ret = JPEG_SaveJPEG_RW(&vars, jpeg_surface, dst, quality);
 
     if (jpeg_surface != surface) {
         SDL_FreeSurface(jpeg_surface);
     }
-
-    result = 0;
-
-done:
-    if (freedst) {
-        SDL_RWclose(dst);
-    }
-    return result;
+    return ret;
 }
 
 #elif defined(USE_STBIMAGE)
@@ -687,7 +697,7 @@ static void IMG_SaveJPG_RW_tinyjpeg_callback(void* context, void* data, int size
     SDL_RWwrite((SDL_RWops*) context, data, 1, size);
 }
 
-static int IMG_SaveJPG_RW_tinyjpeg(SDL_Surface *surface, SDL_RWops *dst, int freedst, int quality)
+static int IMG_SaveJPG_RW_tinyjpeg(SDL_Surface *surface, SDL_RWops *dst, int quality)
 {
     /* The JPEG library reads bytes in R,G,B order, so this is the right
      * encoding for either endianness */
@@ -695,16 +705,11 @@ static int IMG_SaveJPG_RW_tinyjpeg(SDL_Surface *surface, SDL_RWops *dst, int fre
     SDL_Surface* jpeg_surface = surface;
     int result = -1;
 
-    if (!dst) {
-        SDL_SetError("Passed NULL dst");
-        goto done;
-    }
-
     /* Convert surface to format we can save */
     if (surface->format->format != jpg_format) {
         jpeg_surface = SDL_ConvertSurfaceFormat(surface, jpg_format, 0);
         if (!jpeg_surface) {
-            goto done;
+            return -1;
         }
     }
 
@@ -735,11 +740,6 @@ static int IMG_SaveJPG_RW_tinyjpeg(SDL_Surface *surface, SDL_RWops *dst, int fre
     if (result < 0) {
         SDL_SetError("tinyjpeg error");
     }
-
-done:
-    if (freedst) {
-        SDL_RWclose(dst);
-    }
     return result;
 }
 
@@ -757,22 +757,31 @@ int IMG_SaveJPG(SDL_Surface *surface, const char *file, int quality)
 
 int IMG_SaveJPG_RW(SDL_Surface *surface, SDL_RWops *dst, int freedst, int quality)
 {
+    int result = -1;
+
+    if (!dst) {
+        return IMG_SetError("Passed NULL dst");
+    }
+
 #if SDL_IMAGE_SAVE_JPG
 #ifdef USE_JPEGLIB
-    if ((IMG_Init(IMG_INIT_JPG) & IMG_INIT_JPG) != 0) {
-        if (IMG_SaveJPG_RW_jpeglib(surface, dst, freedst, quality) == 0) {
-            return 0;
-        }
+    if (result < 0) {
+        result = IMG_SaveJPG_RW_jpeglib(surface, dst, quality);
     }
 #endif
 
 #if defined(LOAD_JPG_DYNAMIC) || !defined(WANT_JPEGLIB)
-    return IMG_SaveJPG_RW_tinyjpeg(surface, dst, freedst, quality);
-#else
-    return -1;
+    if (result < 0) {
+        result = IMG_SaveJPG_RW_tinyjpeg(surface, dst, quality);
+    }
 #endif
 
 #else
-    return IMG_SetError("SDL_image built without JPEG save support");
+    result = IMG_SetError("SDL_image built without JPEG save support");
 #endif
+
+    if (freedst) {
+        SDL_RWclose(dst);
+    }
+    return result;
 }
